"""Defines heat material specifications"""

from __future__ import annotations

from typing import Union

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.material.tcad.heat import FluidMedium
from tidy3d.components.tcad.boundary.abstract import HeatChargeBC
from tidy3d.constants import (
    ACCELERATION,
    GRAV_ACC,
    HEAT_FLUX,
    HEAT_TRANSFER_COEFF,
    KELVIN,
    MICROMETER,
)


class TemperatureBC(HeatChargeBC):
    """Constant temperature thermal boundary conditions.

    Example
    -------
    >>> import tidy3d as td
    >>> bc = td.TemperatureBC(temperature=300)
    """

    temperature: pd.PositiveFloat = pd.Field(
        title="Temperature",
        description=f"Temperature value in units of {KELVIN}.",
        units=KELVIN,
    )


class HeatFluxBC(HeatChargeBC):
    """Constant flux thermal boundary conditions.

    Example
    -------
    >>> import tidy3d as td
    >>> bc = td.HeatFluxBC(flux=1)
    """

    flux: float = pd.Field(
        title="Heat Flux",
        description=f"Heat flux value in units of {HEAT_FLUX}.",
        units=HEAT_FLUX,
    )


class VerticalNaturalConvectionCoeffModel(Tidy3dBaseModel):
    """
    Specification for natural convection from a vertical plate.

    This class calculates the heat transfer coefficient (h) based on fluid
    properties and an expected temperature difference, then provides these
    values as 'base' and 'exponent' for a generalized heat flux equation
    q = base * (T_surf - T_fluid)^exponent + base * (T_surf - T_fluid).
    """

    medium: FluidMedium = pd.Field(
        default=None,
        title="Interface medium",
        description=(
            "The `FluidMedium` used for the heat transfer coefficient calculation. "
            "If `None`, the fluid is automatically deduced from the interface, which can be defined"
            "by either a `MediumMediumInterface` or a `StructureStructureInterface`."
        ),
    )

    plate_length: pd.NonNegativeFloat = pd.Field(
        title="Plate Characteristic Length",
        description="Characteristic length (L), defined as the height of the vertical plate.",
        units=MICROMETER,
    )

    gravity: pd.NonNegativeFloat = pd.Field(
        default=GRAV_ACC,
        title="Gravitational Acceleration",
        description="Gravitational acceleration (g).",
        units=ACCELERATION,
    )

    def from_si_units(
        plate_length: pd.NonNegativeFloat,
        medium: FluidMedium = None,
        gravity: pd.NonNegativeFloat = GRAV_ACC * 1e-6,
    ):
        """
        Create an instance from standard SI units.

        Args:
            plate_length: Plate characteristic length in [m].
            gravity: Gravitational acceleration in [m/s**2].

        Returns:
            An instance of VerticalNaturalConvectionCoeffModel with all values
            converted to Tidy3D's internal unit system.
        """

        # --- Apply conversion factors ---
        # value_tidy = value_si * factor
        plate_length_tidy = plate_length * 1e6  # m -> um
        g_tidy = gravity * 1e6  # m/s**2 -> um/s**2

        return VerticalNaturalConvectionCoeffModel(
            medium=medium,
            plate_length=plate_length_tidy,
            gravity=g_tidy,
        )


class ConvectionBC(HeatChargeBC):
    """Convective thermal boundary conditions.

    Example
    -------
    >>> import tidy3d as td
    >>> bc = td.ConvectionBC(ambient_temperature=300, transfer_coeff=1)

    >>> # Convection with a natural convection model.
    >>> # First, define the fluid medium (e.g. air at 300 K).
    >>> air = td.FluidMedium.from_si_units(
    ...     thermal_conductivity=0.0257,  # Unit: W/(m*K)
    ...     viscosity=1.81e-5,          # Unit: Pa*s
    ...     specific_heat=1005,         # Unit: J/(kg*K)
    ...     density=1.204,              # Unit: kg/m^3
    ...     expansivity=1/293.15        # Unit: 1/K
    ... )
    >>>
    >>> # Next, create the model, which requires the fluid and a characteristic length.
    >>> natural_conv_model = td.VerticalNaturalConvectionCoeffModel.from_si_units(
    ...     medium=air, plate_length=1e-5
    ... )
    >>>
    >>> # Finally, create the boundary condition using this model.
    >>> bc_natural = td.ConvectionBC(
    ...     ambient_temperature=300, transfer_coeff=natural_conv_model
    ... )

    >>> # If the fluid medium is not provided to the coefficient model, it is automatically retrieved from
    >>> # the interface.
    >>> natural_conv_model = td.VerticalNaturalConvectionCoeffModel.from_si_units(plate_length=1e-5)
    >>> bc_natural_nom = td.ConvectionBC(
    ...     ambient_temperature=300, transfer_coeff=natural_conv_model
    ... )
    """

    ambient_temperature: pd.PositiveFloat = pd.Field(
        title="Ambient Temperature",
        description=f"Ambient temperature value in units of {KELVIN}.",
        units=KELVIN,
    )

    transfer_coeff: Union[pd.NonNegativeFloat, VerticalNaturalConvectionCoeffModel] = pd.Field(
        title="Heat Transfer Coefficient",
        description=f"Heat flux value in units of {HEAT_TRANSFER_COEFF}.",
        units=HEAT_TRANSFER_COEFF,
    )



"""Defines specifications for source frames."""

from __future__ import annotations

from abc import ABC

import pydantic.v1 as pydantic

from tidy3d.components.base import Tidy3dBaseModel


class AbstractSourceFrame(Tidy3dBaseModel, ABC):
    """Abstract base class for all source frames."""

    length: int = pydantic.Field(
        2,
        title="Length",
        description="The length of the frame, specified as the number of cells along the source "
        "injection direction.",
        gt=0,
    )


class PECFrame(AbstractSourceFrame):
    """PEC source frame."""



# define penalties applied to parameters from design region
from __future__ import annotations

import abc
import typing

import autograd.numpy as anp
import pydantic.v1 as pd

from tidy3d.constants import MICROMETER
from tidy3d.plugins.autograd.invdes import make_erosion_dilation_penalty

from .base import InvdesBaseModel


class AbstractPenalty(InvdesBaseModel, abc.ABC):
    """Base class for penalties added to ``invdes.DesignRegion`` objects."""

    weight: pd.NonNegativeFloat = pd.Field(
        1.0,
        title="Weight",
        description="When this penalty is evaluated, it will be weighted by this "
        "value. Note that the optimizer seeks to maximize the objective function and "
        "subtracts the penalty strength times this weight from the objective function. ",
    )

    @abc.abstractmethod
    def evaluate(self) -> float:
        """Evaluate the penalty on supplied values."""

    def __call__(self, *args, **kwargs) -> float:
        return self.evaluate(*args, **kwargs)


class ErosionDilationPenalty(AbstractPenalty):
    """Erosion and dilation penalty, addable to ``tdi.TopologyDesignRegion.penalties``.
    Uses filtering and projection methods to erode and dilate the features within this array.
    Measures the change in the array after eroding and dilating (and also dilating and eroding).
    Returns a penalty proportional to the magnitude of this change.
    The amount of change under dilation and erosion is minimized if the structure has large feature
    sizes and large radius of curvature relative to the length scale.

    Note
    ----
    For more details, refer to chapter 4 of Hammond, A., "High-Efficiency Topology Optimization
    for Very Large-Scale Integrated-Photonics Inverse Design" (2022).

    .. image:: ../../_static/img/erosion_dilation.png

    """

    length_scale: pd.PositiveFloat = pd.Field(
        ...,
        title="Length Scale",
        description="Length scale of erosion and dilation. "
        "Corresponds to ``radius`` in the :class:`ConicFilter` used for filtering. "
        "The parameter array is dilated and eroded by half of this value with each operation. "
        "Roughly corresponds to the desired minimum feature size and radius of curvature.",
        units=MICROMETER,
    )

    beta: float = pd.Field(
        100.0,
        ge=1.0,
        title="Projection Beta",
        description="Strength of the ``tanh`` projection. "
        "Corresponds to ``beta`` in the :class:`BinaryProjector. "
        "Higher values correspond to stronger discretization.",
    )

    eta0: float = pd.Field(
        0.5,
        ge=0.0,
        le=1.0,
        title="Projection Midpoint",
        description="Value between 0 and 1 that sets the projection midpoint. In other words, "
        "for values of ``eta0``, the projected values are halfway between minimum and maximum. "
        "Corresponds to ``eta`` in the :class:`BinaryProjector`.",
    )

    delta_eta: float = pd.Field(
        0.01,
        ge=0.0,
        le=1.0,
        title="Delta Eta Cutoff",
        description="The binarization threshold for erosion and dilation operations "
        "The thresholds are ``0 + delta_eta`` on the low end and ``1 - delta_eta`` on the high end. "
        "The default value balances binarization with differentiability so we strongly suggest "
        "using it unless there is a good reason to set it differently.",
    )

    def evaluate(self, x: anp.ndarray, pixel_size: float) -> float:
        """Evaluate this penalty."""
        penalty_fn = make_erosion_dilation_penalty(
            self.length_scale, pixel_size, beta=self.beta, eta=self.eta0, delta_eta=self.delta_eta
        )
        penalty_unweighted = penalty_fn(x)
        return self.weight * penalty_unweighted


PenaltyType = typing.Union[ErosionDilationPenalty]



"""Defines heat material specifications"""

from __future__ import annotations

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.bc_placement import BCPlacementType
from tidy3d.components.tcad.types import HeatChargeBCType
from tidy3d.components.types import TYPE_TAG_STR


class HeatChargeBoundarySpec(Tidy3dBaseModel):
    """Heat-Charge boundary conditions specification.

    Example
    -------
    >>> import tidy3d as td
    >>> bc_v1 = td.HeatChargeBoundarySpec(
    ...   condition=td.VoltageBC(source=td.DCVoltageSource(voltage=0)),
    ...   placement=td.StructureBoundary(structure="contact_left"),
    ... )
    """

    placement: BCPlacementType = pd.Field(
        title="Boundary Conditions Placement",
        description="Location to apply boundary conditions.",
        discriminator=TYPE_TAG_STR,
    )

    condition: HeatChargeBCType = pd.Field(
        title="Boundary Conditions",
        description="Boundary conditions to apply at the selected location.",
        discriminator=TYPE_TAG_STR,
    )


class HeatBoundarySpec(HeatChargeBoundarySpec):
    """Heat BC specification. DEPRECIATED.

    Warning
    -------
        Included backward-compatibility only.

    Example
    --------
    >>> import tidy3d as td
    >>> bc_spec = td.HeatBoundarySpec(
    ...     placement=td.SimulationBoundary(),
    ...     condition=td.ConvectionBC(ambient_temperature=300, transfer_coeff=1),
    ... )

    """



"""Helpers for automatic setup of path integrals."""

from __future__ import annotations

from tidy3d.components.geometry.base import Box
from tidy3d.components.geometry.utils import (
    SnapBehavior,
    SnapLocation,
    SnappingSpec,
    snap_box_to_grid,
)
from tidy3d.components.grid.grid import Grid
from tidy3d.components.lumped_element import LinearLumpedElement
from tidy3d.components.types import Direction

from .path_integrals import (
    CurrentIntegralAxisAligned,
    VoltageIntegralAxisAligned,
)


def path_integrals_from_lumped_element(
    lumped_element: LinearLumpedElement, grid: Grid, polarity: Direction = "+"
) -> tuple[VoltageIntegralAxisAligned, CurrentIntegralAxisAligned]:
    """Helper to create a :class:`.VoltageIntegralAxisAligned` and :class:`.CurrentIntegralAxisAligned`
    from a supplied :class:`.LinearLumpedElement`. Takes into account any snapping the lumped element
    undergoes using the supplied :class:`.Grid`.

    Parameters
    ----------
    lumped_element : :class:`.LinearLumpedElement`
        Position along the voltage axis of the positive terminal.
    grid : :class:`.Grid`
        Position along the voltage axis of the negative terminal.
    polarity : Direction
        Choice for defining voltage. When positive, the terminal of the lumped element with
        the greatest coordinate is considered the positive terminal.
    Returns
    -------
    VoltageIntegralAxisAligned
        The created path integral for computing voltage between the two terminals of the :class:`.LinearLumpedElement`.
    CurrentIntegralAxisAligned
        The created path integral for computing current flowing through the :class:`.LinearLumpedElement`.
    """

    # Quick access to voltage and the primary current axis
    V_axis = lumped_element.voltage_axis
    I_axis = lumped_element.lateral_axis

    # The exact position of the lumped element after any possible snapping
    lumped_element_box = lumped_element._create_box_for_network(grid=grid)

    V_size = [0, 0, 0]
    V_size[V_axis] = lumped_element_box.size[V_axis]
    voltage_integral = VoltageIntegralAxisAligned(
        center=lumped_element_box.center,
        size=V_size,
        sign=polarity,
        extrapolate_to_endpoints=True,
        snap_path_to_grid=True,
    )

    # Snap the current integral to a box that encloses the element along the lateral and normal axes
    # using the closest positions of the magnetic field
    snap_location = [SnapLocation.Center] * 3
    snap_behavior = [SnapBehavior.Expand] * 3
    # Don't need to snap along voltage axis, since it will already be snapped from the lumped element's box
    snap_behavior[V_axis] = SnapBehavior.Off
    snap_spec = SnappingSpec(location=snap_location, behavior=snap_behavior)

    I_size = [0, 0, 0]
    I_size[I_axis] = lumped_element_box.size[I_axis]
    current_box = Box(center=lumped_element_box.center, size=I_size)
    current_box = snap_box_to_grid(grid, current_box, snap_spec)
    # Convention is current flows from plus to minus terminals
    current_sign = "-" if polarity == "+" else "+"
    current_integral = CurrentIntegralAxisAligned(
        center=current_box.center,
        size=current_box.size,
        sign=current_sign,
        snap_contour_to_grid=True,
        extrapolate_to_endpoints=True,
    )

    return (voltage_integral, current_integral)



from __future__ import annotations

from typing import Literal

import pydantic.v1 as pd

# S matrix indices and entries for the ModalComponentModeler
MatrixIndex = tuple[str, pd.NonNegativeInt]  # the 'i' in S_ij
Element = tuple[MatrixIndex, MatrixIndex]  # the 'ij' in S_ij
# S matrix indices and entries for the TerminalComponentModeler
NetworkIndex = str  # the 'i' in S_ij
NetworkElement = tuple[NetworkIndex, NetworkIndex]  # the 'ij' in S_ij

# The definition of wave amplitudes used to construct the scattering matrix
# in the TerminalComponentModeler
SParamDef = Literal["pseudo", "power"]



from __future__ import annotations

from tidy3d.log import log

_ORIGINAL_PARAMS = None


def apply_tidy3d_params():
    """
    Applies a set of defaults to the matplotlib params that are following the tidy3d color palettes and design.
    """
    global _ORIGINAL_PARAMS
    try:
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        _ORIGINAL_PARAMS = mpl.rcParams.copy()

        try:
            plt.style.use("tidy3d.style")
        except Exception as e:
            log.error(f"Failed to apply Tidy3D plotting style on import. Error: {e}")
            _ORIGINAL_PARAMS = {}
    except ImportError:
        pass


def restore_matplotlib_rcparams():
    """
    Resets matplotlib rcParams to the values they had before the Tidy3D
    style was automatically applied on import.
    """
    global _ORIGINAL_PARAMS
    try:
        import matplotlib.pyplot as plt
        from matplotlib import style

        if not _ORIGINAL_PARAMS:
            style.use("default")
            return

        plt.rcParams.update(_ORIGINAL_PARAMS)
    except ImportError:
        log.error("Matplotlib is not installed on your system. Failed to reset to default styles.")
    except Exception as e:
        log.error(f"Failed to reset previous Matplotlib style. Error: {e}")



from __future__ import annotations

from typing import Any, Optional

import pydantic.v1 as pd

from .base import Expression
from .types import NumberType


class Variable(Expression):
    """
    Variable class represents a placeholder for a value provided at evaluation time.

    Attributes
    ----------
    name : Optional[str] = None
        The name of the variable used for lookup during evaluation.

    Methods
    -------
    evaluate(*args, **kwargs)
        Evaluates the variable by retrieving its value from provided arguments.

    Notes
    -----
    - If `name` is `None`, the variable expects a single positional argument during evaluation.
    - If `name` is provided, the variable expects a corresponding keyword argument during evaluation.
    - Mixing positional and keyword arguments is allowed.
    - Multiple positional arguments are disallowed and will raise a `ValueError`.

    Examples
    --------
    >>> x = Variable()  # unnamed
    >>> y = Variable(name='y')
    >>> z = Variable(name='z')
    >>> expr = x + y + z
    >>> expr(5, y=3, z=2)
    10
    """

    name: Optional[str] = pd.Field(
        None,
        title="Name",
        description="The name of the variable used for lookup during evaluation.",
    )

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        if self.name:
            if self.name not in kwargs:
                raise ValueError(f"Variable '{self.name}' not provided.")
            return kwargs[self.name]
        if not args:
            raise ValueError("No positional argument provided for unnamed variable.")
        if len(args) > 1:
            raise ValueError("Multiple positional arguments provided for unnamed variable.")
        return args[0]

    def __repr__(self) -> str:
        return self.name if self.name else "Variable()"


class Constant(Variable):
    """
    Constant class represents a fixed value in an expression.

    Attributes
    ----------
    value : NumberType
        The fixed value of the constant.

    Methods
    -------
    evaluate(*args, **kwargs)
        Returns the value of the constant.

    Examples
    --------
    >>> c = Constant(5)
    >>> c.evaluate()
    5
    """

    value: NumberType = pd.Field(
        ...,
        title="Value",
        description="The fixed value of the constant.",
    )

    def __init__(self, value: NumberType, **kwargs: dict[str, Any]) -> None:
        super().__init__(value=value, **kwargs)

    def evaluate(self, *args: Any, **kwargs: Any) -> NumberType:
        return self.value

    def __repr__(self) -> str:
        return f"{self.value}"



"""Monitor level data, store the DataArrays associated with a single heat-charge monitor."""

from __future__ import annotations

from typing import Optional, Union

import pydantic.v1 as pd

from tidy3d.components.data.data_array import (
    DataArray,
    ScalarFieldTimeDataArray,
    SpatialDataArray,
)
from tidy3d.components.data.utils import TetrahedralGridDataset, TriangularGridDataset
from tidy3d.components.tcad.data.monitor_data.abstract import HeatChargeMonitorData
from tidy3d.components.tcad.monitors.heat import (
    TemperatureMonitor,
)
from tidy3d.components.types import annotate_type
from tidy3d.constants import KELVIN

FieldDataset = Union[
    SpatialDataArray,
    ScalarFieldTimeDataArray,
    annotate_type(Union[TriangularGridDataset, TetrahedralGridDataset]),
]
UnstructuredFieldType = Union[TriangularGridDataset, TetrahedralGridDataset]


class TemperatureData(HeatChargeMonitorData):
    """Data associated with a :class:`TemperatureMonitor`: spatial temperature field.

    Example
    -------
    >>> from tidy3d import TemperatureMonitor, SpatialDataArray
    >>> import numpy as np
    >>> temp_data = SpatialDataArray(
    ...     np.ones((2, 3, 4)), coords={"x": [0, 1], "y": [0, 1, 2], "z": [0, 1, 2, 3]}
    ... )
    >>> temp_mnt = TemperatureMonitor(size=(1, 2, 3), name="temperature")
    >>> temp_mnt_data = TemperatureData(
    ...     monitor=temp_mnt, temperature=temp_data, symmetry=(0, 1, 0), symmetry_center=(0, 0, 0)
    ... )
    >>> temp_mnt_data_expanded = temp_mnt_data.symmetry_expanded_copy
    """

    monitor: TemperatureMonitor = pd.Field(
        ..., title="Monitor", description="Temperature monitor associated with the data."
    )

    temperature: Optional[FieldDataset] = pd.Field(
        ...,
        title="Temperature",
        description="Spatial temperature field.",
        units=KELVIN,
    )

    @property
    def field_components(self) -> dict[str, DataArray]:
        """Maps the field components to their associated data."""
        return {"temperature": self.temperature}



"""Objects that define how data is recorded from simulation."""

from __future__ import annotations

from pydantic.v1 import Field, PositiveInt

from tidy3d.components.tcad.monitors.abstract import HeatChargeMonitor


class TemperatureMonitor(HeatChargeMonitor):
    """Temperature monitor."""

    interval: PositiveInt = Field(
        1,
        title="Interval",
        description="Sampling rate of the monitor: number of time steps between each measurement. "
        "Set ``interval`` to 1 for the highest possible resolution in time. "
        "Higher integer values down-sample the data by measuring every ``interval`` time steps. "
        "This can be useful for reducing data storage as needed by the application."
        "NOTE: this is only relevant for unsteady (transient) Heat simulations. ",
    )



"""File containing classes required for the setup of a DEVSIM case."""

from __future__ import annotations

from typing import Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.geometry.base import Box
from tidy3d.constants import PERCMCUBE
from tidy3d.exceptions import SetupError


class AbstractDopingBox(Box):
    """Derived class from Box to deal with dopings"""

    def _normal_dim(self):
        """Returns the normal direction if the box is 2D. False otherwise"""

        normal_dim = None
        for dim in range(3):
            if self.size[dim] == np.inf:
                if normal_dim is not None:
                    raise SetupError("Only 3D and 2D boxes are considered for doping.")
                normal_dim = dim

        return normal_dim

    def _get_indices_in_box(self, coords: dict, meshgrid: bool = True):
        """Returns locations inside box"""

        # work out whether x,y, and z are present
        dim_missing = len(list(coords.keys())) < 3
        if dim_missing:
            for var_name in "xyz":
                if var_name not in coords:
                    coords[var_name] = [0]

        # work out whether the dimensions are 2D
        normal_axis = None
        # normal_position = None
        for dim in range(3):
            var_name = "xyz"[dim]
            if len(coords[var_name]) == 1:
                normal_axis = dim
                # normal_position = coords[var_name][0]

        if all(len(coords[var_name]) == 1 for var_name in "xyz"):
            # if all coordinates have 1 point, we don't assume  2D unless the box itself is.
            normal_axis = None

        # if provided coordinates are 3D, check if box is 2D
        if normal_axis is None:
            normal_axis = self._normal_dim()

        if meshgrid:
            X, Y, Z = np.meshgrid(coords["x"], coords["y"], coords["z"], indexing="ij")
        else:
            X = coords["x"]
            Y = coords["y"]
            Z = coords["z"]

        new_bounds = [list(self.bounds[0]), list(self.bounds[1])]
        for d in range(3):
            if new_bounds[0][d] == new_bounds[1][d]:
                new_bounds[0][d] = -np.inf
                new_bounds[1][d] = np.inf

        # let's assume some of these coordinates may lay outside the box
        indices_in_box = np.logical_and(new_bounds[0][0] <= X, new_bounds[1][0] >= X)
        indices_in_box = np.logical_and(indices_in_box, new_bounds[0][1] <= Y)
        indices_in_box = np.logical_and(indices_in_box, new_bounds[1][1] >= Y)
        indices_in_box = np.logical_and(indices_in_box, new_bounds[0][2] <= Z)
        indices_in_box = np.logical_and(indices_in_box, new_bounds[1][2] >= Z)

        return indices_in_box, X, Y, Z, normal_axis

    @pd.root_validator(skip_on_failure=True)
    def check_dimensions(cls, values):
        """Make sure dimensionality is specified correctly. I.e.,
        a 2D box must be defined with an inf size in the normal direction."""

        size = values["size"]
        for dim in range(3):
            if size[dim] == 0:
                zero_dim_name = "xyz"[dim]

                raise SetupError(
                    f"The doping box has been set up with 0 size in the {zero_dim_name} direction. "
                    "If this was intended to be translationally invariant, the box must have a large "
                    "or infinite ('td.inf') size in the perpendicular direction."
                )

        return values


class ConstantDoping(AbstractDopingBox):
    """
    Sets constant doping :math:`N` in the specified box with a :parameter`size` and :parameter:`concentration`.

    For translationally invariant behavior in one dimension, the box must have infinite size in the
    homogenous (invariant) direction.

    Example
    -------
    >>> import tidy3d as td
    >>> box_coords = [
    ...     [-1, -1, -1],
    ...     [1, 1, 1]
    ... ]
    >>> constant_box1 = td.ConstantDoping(center=(0, 0, 0), size=(2, 2, 2), concentration=1e18)
    >>> constant_box2 = td.ConstantDoping.from_bounds(rmin=box_coords[0], rmax=box_coords[1], concentration=1e18)
    """

    concentration: pd.NonNegativeFloat = pd.Field(
        default=0,
        title="Doping concentration density.",
        description="Doping concentration density in #/cm^3.",
        units=PERCMCUBE,
    )

    def _get_contrib(self, coords: dict, meshgrid: bool = True):
        """Returns the contribution to the doping a the locations specified in coords"""

        indices_in_box, X, _, _, normal_axis = self._get_indices_in_box(
            coords=coords, meshgrid=meshgrid
        )

        contrib = np.zeros(X.shape)
        contrib[indices_in_box] = self.concentration

        if normal_axis is not None and meshgrid:
            slices = [slice(None)] * X.ndim
            slices[normal_axis] = 0
            return contrib[tuple(slices)]
        return contrib


class GaussianDoping(AbstractDopingBox):
    """Sets a gaussian doping in the specified box.

    For translationally invariant behavior in one dimension, the box must have infinite size in the
    homogenous (invariant) direction.

    Notes
    -----
    The Gaussian doping concentration :math:`N` is defined in the following manner:

    - :math:`N=N_{\\text{max}}` at locations more than :math:``width`` um away from the sides of the box.
    - :math:`N=N_{\\text{ref}}` at location on the box sides.
    - a Gaussian variation between :math:`N_{\\text{max}}` and  :math:`N_{\\text{ref}}`  at locations less than ``width``
    um away from the sides.

    By definition, all sides of the box will have concentration :math:`N_{\\text{ref}}` (except the side specified
    as source) and the center of the box (``width`` away from the box sides) will have a concentration
    :math:`N_{\\text{max}}`.

    .. math::

        N = \\{N_{\\text{max}}\\} \\exp \\left[
        - \\ln \\left( \\frac{\\{N_{\\text{max}}\\}}{\\{N_{\\text{ref}}\\}} \\right)
        \\left( \\frac{(x|y|z) - \\{(x|y|z)_{\\text{box}}\\}}{\\text{width}} \\right)^2
        \\right]

    Example
    -------
    >>> import tidy3d as td
    >>> box_coords = [
    ...     [-1, -1, -1],
    ...     [1, 1, 1]
    ... ]
    >>> gaussian_box1 = td.GaussianDoping(
    ...     center=(0, 0, 0),
    ...     size=(2, 2, 2),
    ...     ref_con=1e15,
    ...     concentration=1e18,
    ...     width=0.1,
    ...     source="xmin"
    ... )
    >>> gaussian_box2 = td.GaussianDoping.from_bounds(
    ...     rmin=box_coords[0],
    ...     rmax=box_coords[1],
    ...     ref_con=1e15,
    ...     concentration=1e18,
    ...     width=0.1,
    ...     source="xmin"
    ... )
    """

    ref_con: pd.PositiveFloat = pd.Field(
        title="Reference concentration.",
        description="Reference concentration. This is the minimum concentration in the box "
        "and it is attained at the edges/faces of the box.",
    )

    concentration: pd.PositiveFloat = pd.Field(
        title="Concentration",
        description="The concentration at the center of the box.",
    )

    width: pd.PositiveFloat = pd.Field(
        title="Width of the gaussian.",
        description="Width of the gaussian. The concentration will transition from "
        "'concentration' at the center of the box to 'ref_con' at the edge/face "
        "of the box in a distance equal to 'width'. ",
    )

    source: str = pd.Field(
        "xmin",
        title="Source face",
        description="Specifies the side of the box acting as the source, i.e., "
        "the face specified does not have a gaussian evolution normal to it, instead "
        "the concentration is constant from this face. Accepted values for 'source' "
        "are ['xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax']",
    )

    @cached_property
    def sigma(self):
        """The sigma parameter of the pseudo-gaussian"""
        return np.sqrt(-self.width * self.width / 2 / np.log(self.ref_con / self.concentration))

    def _get_contrib(self, coords: dict, meshgrid: bool = True):
        """Returns the contribution to the doping a the locations specified in coords"""

        indices_in_box, X, Y, Z, normal_axis = self._get_indices_in_box(
            coords=coords, meshgrid=meshgrid
        )

        x_contrib = np.ones(X.shape)
        if normal_axis != 0:
            x_contrib = np.zeros(X.shape)
            x_contrib[indices_in_box] = 1
            # lower x face
            if self.source != "xmin":
                x0 = self.bounds[0][0]
                indices = np.logical_and(x0 <= X, x0 + self.width >= X)
                indices = np.logical_and(indices, indices_in_box)
                x_contrib[indices] = np.exp(
                    -(X[indices] - x0 - self.width)
                    * (X[indices] - x0 - self.width)
                    / 2
                    / self.sigma
                    / self.sigma
                )
            # higher x face
            if self.source != "xmax":
                x1 = self.bounds[1][0]
                indices = np.logical_and(x1 - self.width <= X, x1 >= X)
                indices = np.logical_and(indices, indices_in_box)
                x_contrib[indices] = np.exp(
                    -(X[indices] - x1 + self.width)
                    * (X[indices] - x1 + self.width)
                    / 2
                    / self.sigma
                    / self.sigma
                )

        y_contrib = np.ones(X.shape)
        if normal_axis != 1:
            y_contrib = np.zeros(X.shape)
            y_contrib[indices_in_box] = 1
            # lower y face
            if self.source != "ymin":
                y0 = self.bounds[0][1]
                indices = np.logical_and(y0 <= Y, y0 + self.width >= Y)
                indices = np.logical_and(indices, indices_in_box)
                y_contrib[indices] = np.exp(
                    -(Y[indices] - y0 - self.width)
                    * (Y[indices] - y0 - self.width)
                    / 2
                    / self.sigma
                    / self.sigma
                )
            # higher y face
            if self.source != "ymax":
                y1 = self.bounds[1][1]
                indices = np.logical_and(y1 - self.width <= Y, y1 >= Y)
                indices = np.logical_and(indices, indices_in_box)
                y_contrib[indices] = np.exp(
                    -(Y[indices] - y1 + self.width)
                    * (Y[indices] - y1 + self.width)
                    / 2
                    / self.sigma
                    / self.sigma
                )

        z_contrib = np.ones(X.shape)
        if normal_axis != 2:
            z_contrib = np.zeros(X.shape)
            z_contrib[indices_in_box] = 1
            # lower z face
            if self.source != "zmin":
                z0 = self.bounds[0][2]
                indices = np.logical_and(z0 <= Z, z0 + self.width >= Z)
                indices = np.logical_and(indices, indices_in_box)
                z_contrib[indices] = np.exp(
                    -(Z[indices] - z0 - self.width)
                    * (Z[indices] - z0 - self.width)
                    / 2
                    / self.sigma
                    / self.sigma
                )
            # higher z face
            if self.source != "zmax":
                z1 = self.bounds[1][2]
                indices = np.logical_and(z1 - self.width <= Z, z1 >= Z)
                indices = np.logical_and(indices, indices_in_box)
                z_contrib[indices] = np.exp(
                    -(Z[indices] - z1 + self.width)
                    * (Z[indices] - z1 + self.width)
                    / 2
                    / self.sigma
                    / self.sigma
                )

        total_contrib = x_contrib * y_contrib * z_contrib * self.concentration

        if normal_axis is not None and meshgrid:
            slices = [slice(None)] * X.ndim
            slices[normal_axis] = 0
            return total_contrib[tuple(slices)]
        return total_contrib


DopingBoxType = Union[ConstantDoping, GaussianDoping]



"""Defines information about a task"""

from __future__ import annotations

from abc import ABC
from datetime import datetime
from enum import Enum
from typing import Optional

import pydantic.v1 as pydantic


class TaskStatus(Enum):
    """The statuses that the task can be in."""

    INIT = "initialized"
    """The task has been initialized."""

    QUEUE = "queued"
    """The task is in the queue."""

    PRE = "preprocessing"
    """The task is in the preprocessing stage."""

    RUN = "running"
    """The task is running."""

    POST = "postprocessing"
    """The task is in the postprocessing stage."""

    SUCCESS = "success"
    """The task has completed successfully."""

    ERROR = "error"
    """The task has completed with an error."""


class TaskBase(pydantic.BaseModel, ABC):
    """Base configuration for all task objects."""

    class Config:
        """Configuration for TaskBase"""

        arbitrary_types_allowed = True
        """Allow arbitrary types to be used within the model."""


class ChargeType(str, Enum):
    """The payment method of the task."""

    FREE = "free"
    """No payment required."""

    PAID = "paid"
    """Payment required."""


class TaskBlockInfo(TaskBase):
    """Information about the task's block status.

    This includes details about how the task can be blocked by various features
    such as user limits and insufficient balance.
    """

    chargeType: ChargeType = None
    """The type of charge applicable to the task (free or paid)."""

    maxFreeCount: int = None
    """The maximum number of free tasks allowed."""

    maxGridPoints: int = None
    """The maximum number of grid points permitted."""

    maxTimeSteps: int = None
    """The maximum number of time steps allowed."""


class TaskInfo(TaskBase):
    """General information about a task."""

    taskId: str
    """Unique identifier for the task."""

    taskName: str = None
    """Name of the task."""

    nodeSize: int = None
    """Size of the node allocated for the task."""

    completedAt: Optional[datetime] = None
    """Timestamp when the task was completed."""

    status: str = None
    """Current status of the task."""

    realCost: float = None
    """Actual cost incurred by the task."""

    timeSteps: int = None
    """Number of time steps involved in the task."""

    solverVersion: str = None
    """Version of the solver used for the task."""

    createAt: Optional[datetime] = None
    """Timestamp when the task was created."""

    estCostMin: float = None
    """Estimated minimum cost for the task."""

    estCostMax: float = None
    """Estimated maximum cost for the task."""

    realFlexUnit: float = None
    """Actual flexible units used by the task."""

    oriRealFlexUnit: float = None
    """Original real flexible units."""

    estFlexUnit: float = None
    """Estimated flexible units for the task."""

    estFlexCreditTimeStepping: float = None
    """Estimated flexible credits for time stepping."""

    estFlexCreditPostProcess: float = None
    """Estimated flexible credits for post-processing."""

    estFlexCreditMode: float = None
    """Estimated flexible credits based on the mode."""

    s3Storage: float = None
    """Amount of S3 storage used by the task."""

    startSolverTime: Optional[datetime] = None
    """Timestamp when the solver started."""

    finishSolverTime: Optional[datetime] = None
    """Timestamp when the solver finished."""

    totalSolverTime: int = None
    """Total time taken by the solver."""

    callbackUrl: str = None
    """Callback URL for task notifications."""

    taskType: str = None
    """Type of the task."""

    metadataStatus: str = None
    """Status of the metadata for the task."""

    taskBlockInfo: TaskBlockInfo = None
    """Blocking information for the task."""


class RunInfo(TaskBase):
    """Information about the run of a task."""

    perc_done: pydantic.confloat(ge=0.0, le=100.0)
    """Percentage of the task that is completed (0 to 100)."""

    field_decay: pydantic.confloat(ge=0.0, le=1.0)
    """Field decay from the maximum value (0 to 1)."""

    def display(self):
        """Print some info about the task's progress."""
        print(f" - {self.perc_done:.2f} (%) done")
        print(f" - {self.field_decay:.2e} field decay from max")


# ---------------------- Batch (Modeler) detail schema ---------------------- #


class BatchStatus(str, Enum):
    """Enumerates the possible statuses for a batch of tasks."""

    draft = "draft"
    """The batch is being configured and has not been submitted."""
    preprocess = "preprocess"
    """The batch is undergoing preprocessing."""
    validating = "validating"
    """The tasks within the batch are being validated."""
    validate_success = "validate_success"
    """All tasks in the batch passed validation."""
    validate_warn = "validate_warn"
    """Validation passed, but with warnings."""
    validate_fail = "validate_fail"
    """Validation failed for one or more tasks."""
    blocked = "blocked"
    """The batch is blocked and cannot run."""
    running = "running"
    """The batch is currently executing."""
    aborting = "aborting"
    """The batch is in the process of being aborted."""
    run_success = "run_success"
    """The batch completed successfully."""
    postprocess = "postprocess"
    """The batch is undergoing postprocessing."""
    run_failed = "run_failed"
    """The batch execution failed."""
    diverged = "diverged"
    """The simulation in the batch diverged."""
    aborted = "aborted"
    """The batch was successfully aborted."""
    error = "error"
    """An error occurred during the solver run."""


class BatchTaskBlockInfo(TaskBlockInfo):
    """
    Extends `TaskBlockInfo` with specific details for batch task blocking.

    Attributes:
        accountLimit: A usage or cost limit imposed by the user's account.
        taskBlockMsg: A human-readable message describing the reason for the block.
        taskBlockType: The specific type of block (e.g., 'balance', 'limit').
        blockStatus: The current blocking status for the batch.
        taskStatus: The status of the task when it was blocked.
    """

    accountLimit: float = None
    taskBlockMsg: str = None
    taskBlockType: str = None
    blockStatus: str = None
    taskStatus: str = None


class BatchMember(TaskBase):
    """
    Represents a single task within a larger batch operation.

    Attributes:
        refId: A reference identifier for the member task.
        folderId: The identifier of the folder containing the task.
        sweepId: The identifier for the parameter sweep, if applicable.
        taskId: The unique identifier of the task.
        linkedTaskId: The identifier of a task linked to this one.
        groupId: The identifier of the group this task belongs to.
        taskName: The name of the individual task.
        status: The current status of this specific task.
        sweepData: Data associated with a parameter sweep.
        validateInfo: Information related to the task's validation.
        replaceData: Data used for replacements or modifications.
        protocolVersion: The version of the protocol used.
        variable: The variable parameter for this task in a sweep.
        createdAt: The timestamp when the member task was created.
        updatedAt: The timestamp when the member task was last updated.
        denormalizeStatus: The status of the data denormalization process.
        summary: A dictionary containing summary information for the task.
    """

    refId: str = None
    folderId: str = None
    sweepId: str = None
    taskId: str = None
    linkedTaskId: str = None
    groupId: str = None
    taskName: str = None
    status: str = None
    sweepData: str = None
    validateInfo: str = None
    replaceData: str = None
    protocolVersion: str = None
    variable: str = None
    createdAt: Optional[datetime] = None
    updatedAt: Optional[datetime] = None
    denormalizeStatus: str = None
    summary: dict = None


class BatchDetail(TaskBase):
    """
    Provides a detailed, top-level view of a batch of tasks.

    This model serves as the main payload for retrieving comprehensive
    information about a batch operation.

    Attributes:
        refId: A reference identifier for the entire batch.
        optimizationId: Identifier for the optimization process, if any.
        groupId: Identifier for the group the batch belongs to.
        name: The user-defined name of the batch.
        status: The current status of the batch.
        totalStatus: The overall status, consolidating individual task statuses.
        totalTask: The total number of tasks in the batch.
        preprocessSuccess: The count of tasks that completed preprocessing.
        postprocessStatus: The status of the batch's postprocessing stage.
        validateSuccess: The count of tasks that passed validation.
        runSuccess: The count of tasks that ran successfully.
        postprocessSuccess: The count of tasks that completed postprocessing.
        taskBlockInfo: Information on what might be blocking the batch.
        estFlexUnit: The estimated total flexible compute units for the batch.
        totalSeconds: The total time in seconds the batch has taken.
        totalCheckMillis: Total time in milliseconds spent on checks.
        message: A general message providing information about the batch status.
        tasks: A list of `BatchMember` objects, one for each task in the batch.
    """

    refId: str = None
    optimizationId: str = None
    groupId: str = None
    name: str = None
    status: str = None
    totalStatus: BatchStatus = None
    totalTask: int = 0
    preprocessSuccess: int = 0
    postprocessStatus: str = None
    validateSuccess: int = 0
    runSuccess: int = 0
    postprocessSuccess: int = 0
    taskBlockInfo: BatchTaskBlockInfo = None
    estFlexUnit: float = None
    realFlexUnit: float = None
    totalSeconds: int = None
    totalCheckMillis: int = None
    message: str = None
    tasks: list[BatchMember] = []



"""Defines specification for mode solver."""

from __future__ import annotations

from math import isclose
from typing import Literal, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.constants import GLANCING_CUTOFF, MICROMETER, RADIAN, fp_eps
from tidy3d.exceptions import SetupError, ValidationError
from tidy3d.log import log

from .base import Tidy3dBaseModel, skip_if_fields_missing
from .types import Axis2D, TrackFreq

GROUP_INDEX_STEP = 0.005


class ModeSpec(Tidy3dBaseModel):
    """
    Stores specifications for the mode solver to find an electromagntic mode.

    Notes
    -----

        The :attr:`angle_theta` and :attr:`angle_phi` parameters define the injection axis as illustrated in the figure
        below, with respect to the axis normal to the mode plane (``x`` in the figure). Note that :attr:`angle_theta`
        must be smaller than :math:`\\frac{pi}{2}`. To inject in the backward direction, we can still use the
        ``direction`` parameter as also shown in the figure. Similarly, the mode amplitudes computed in mode monitors
        are defined w.r.t. the ``forward`` and ``backward`` directions as illustrated. Note, the planar axes are
        found by popping the injection axis from ``{x,y,z}``. For example, if injection axis is ``y``, the planar
        axes are ordered ``{x,z}``.

        .. image:: ../../notebooks/img/ring_modes.png

        The :attr:`bend_axis` is the axis normal to the plane in which the bend lies, (``z`` in the diagram below). In
        the mode specification, it is defined locally for the mode plane as one of the two axes tangential to the
        plane. In the case of bends that lie in the ``xy``-plane, the mode plane would be either in ``xz`` or in
        ``yz``, so in both cases the correct setting is ``bend_axis=1``, selecting the global ``z``. The
        ``bend_radius`` is counted from the center of the mode plane to the center of the curvature,
        along the tangential axis perpendicular to the bend axis. This radius can also be negative, if the center of
        the mode plane is smaller than the center of the bend.

        .. image:: ../../notebooks/img/mode_angled.png

    Example
    -------
    >>> mode_spec = ModeSpec(num_modes=3, target_neff=1.5)

    See Also
    --------

    **Notebooks**:
        * `Introduction on tidy3d working principles <../../notebooks/Primer.html#Modes>`_
        * `Defining mode sources and monitors <../../notebooks/ModalSourcesMonitors.html>`_
        * `Injecting modes in bent and angled waveguides <../../notebooks/ModesBentAngled.html>`_
        * `Waveguide to ring coupling <../../notebooks/WaveguideToRingCoupling.html>`_

    """

    num_modes: pd.PositiveInt = pd.Field(
        1, title="Number of modes", description="Number of modes returned by mode solver."
    )

    target_neff: pd.PositiveFloat = pd.Field(
        None, title="Target effective index", description="Guess for effective index of the mode."
    )

    num_pml: tuple[pd.NonNegativeInt, pd.NonNegativeInt] = pd.Field(
        (0, 0),
        title="Number of PML layers",
        description="Number of standard pml layers to add in the two tangential axes.",
    )

    filter_pol: Literal["te", "tm"] = pd.Field(
        None,
        title="Polarization filtering",
        description="The solver always computes the ``num_modes`` modes closest to the given "
        "``target_neff``. If ``filter_pol==None``, they are simply sorted in order of decreasing "
        "effective index. If a polarization filter is selected, the modes are rearranged such that "
        "the first ``n_pol`` modes in the list are the ones with the selected polarization "
        "fraction larger than or equal to 0.5, while the next ``num_modes - n_pol`` modes are the "
        "ones where it is smaller than 0.5 (i.e. the opposite polarization fraction is larger than "
        "0.5). Within each polarization subset, the modes are still ordered by decreasing "
        "effective index. "
        "``te``-fraction is defined as the integrated intensity of the E-field component parallel "
        "to the first plane axis, normalized to the total in-plane E-field intensity. Conversely, "
        "``tm``-fraction uses the E field component parallel to the second plane axis.",
    )

    angle_theta: float = pd.Field(
        0.0,
        title="Polar Angle",
        description="Polar angle of the propagation axis from the injection axis.",
        units=RADIAN,
    )

    angle_phi: float = pd.Field(
        0.0,
        title="Azimuth Angle",
        description="Azimuth angle of the propagation axis in the plane orthogonal to the "
        "injection axis.",
        units=RADIAN,
    )

    precision: Literal["auto", "single", "double"] = pd.Field(
        "double",
        title="single, double, or automatic precision in mode solver",
        description="The solver will be faster and using less memory under "
        "single precision, but more accurate under double precision. "
        "Choose ``'auto'`` to apply double precision if the simulation contains a good "
        "conductor, single precision otherwise.",
    )

    bend_radius: float = pd.Field(
        None,
        title="Bend radius",
        description="A curvature radius for simulation of waveguide bends. Can be negative, in "
        "which case the mode plane center has a smaller value than the curvature center along the "
        "tangential axis perpendicular to the bend axis.",
        units=MICROMETER,
    )

    bend_axis: Axis2D = pd.Field(
        None,
        title="Bend axis",
        description="Index into the two tangential axes defining the normal to the "
        "plane in which the bend lies. This must be provided if ``bend_radius`` is not ``None``. "
        "For example, for a ring in the global xy-plane, and a mode plane in either the xz or the "
        "yz plane, the ``bend_axis`` is always 1 (the global z axis).",
    )

    angle_rotation: bool = pd.Field(
        False,
        title="Use fields rotation when angle_theta is not zero",
        description="Defines how modes are computed when angle_theta is not zero. "
        "If 'False', a coordinate transformation is applied through the permittivity and permeability tensors."
        "If 'True', the structures in the simulation are first rotated to compute a mode solution at "
        "a reference plane normal to the structure's azimuthal direction. Then, the fields are rotated "
        "to align with the mode plane, using the 'n_eff' calculated at the reference plane. The second option can "
        "produce more accurate results, but more care must be taken, for example, in ensuring that the "
        "original mode plane intersects the correct geometries in the simulation with rotated structures. "
        "Note: currently only supported when 'angle_phi' is a multiple of 'np.pi'.",
    )

    track_freq: Union[TrackFreq, None] = pd.Field(
        "central",
        title="Mode Tracking Frequency",
        description="Parameter that turns on/off mode tracking based on their similarity. "
        "Can take values ``'lowest'``, ``'central'``, or ``'highest'``, which correspond to "
        "mode tracking based on the lowest, central, or highest frequency. "
        "If ``None`` no mode tracking is performed.",
    )

    group_index_step: Union[pd.PositiveFloat, bool] = pd.Field(
        False,
        title="Frequency step for group index computation",
        description="Control the computation of the group index alongside the effective index. If "
        "set to a positive value, it sets the fractional frequency step used in the numerical "
        "differentiation of the effective index to compute the group index. If set to `True`, the "
        f"default of {GROUP_INDEX_STEP} is used.",
    )

    @pd.validator("bend_axis", always=True)
    @skip_if_fields_missing(["bend_radius"])
    def bend_axis_given(cls, val, values):
        """Check that ``bend_axis`` is provided if ``bend_radius`` is not ``None``"""
        if val is None and values.get("bend_radius") is not None:
            raise SetupError("'bend_axis' must also be defined if 'bend_radius' is defined.")
        return val

    @pd.validator("bend_radius", always=True)
    def bend_radius_not_zero(cls, val, values):
        """Check that ``bend_raidus`` magnitude is not close to zero.`"""
        if val is not None and isclose(val, 0):
            raise SetupError("The magnitude of 'bend_radius' must be larger than 0.")
        return val

    @pd.validator("angle_theta", allow_reuse=True, always=True)
    def glancing_incidence(cls, val):
        """Warn if close to glancing incidence."""
        if np.abs(np.pi / 2 - val) < GLANCING_CUTOFF:
            raise SetupError(
                "Mode propagation axis too close to glancing angle for accurate injection. "
                "For best results, switch the injection axis."
            )
        return val

    # Must be executed before type validation by pydantic, otherwise True is converted to 1.0
    @pd.validator("group_index_step", pre=True)
    def assign_default_on_true(cls, val):
        """Assign the default fractional frequency step value if not provided."""
        if val is True:
            return GROUP_INDEX_STEP
        return val

    @pd.validator("group_index_step")
    def check_group_step_size(cls, val):
        """Ensure a reasonable group index step is used."""
        if val >= 1:
            raise ValidationError(
                "Parameter 'group_index_step' is a fractional value. It must be less than 1."
            )
        return val

    @pd.root_validator(skip_on_failure=True)
    def check_precision(cls, values):
        """Verify critical ModeSpec settings for group index calculation."""
        if values["group_index_step"] > 0:
            if values["track_freq"] is None:
                log.warning(
                    "Group index calculation without mode tracking can lead to incorrect results "
                    "around mode crossings. Consider setting 'track_freq' to 'central'."
                )

            # multiply by 5 to be safe
            if values["group_index_step"] < 5 * fp_eps and values["precision"] != "double":
                log.warning(
                    "Group index step is too small! "
                    "The results might be fully corrupted by numerical errors. "
                    "For more accurate results, please consider using 'double' precision, "
                    "or increasing the value of 'group_index_step'."
                )

        return values

    @pd.validator("angle_rotation")
    def angle_rotation_with_phi(cls, val, values):
        """Currently ``angle_rotation`` is only supported with ``angle_phi % (np.pi / 2) == 0``."""
        if val and not isclose(values["angle_phi"] % (np.pi / 2), 0):
            raise ValidationError(
                "Parameter 'angle_phi' must be a multiple of 'np.pi / 2' when 'angle_rotation' is "
                "enabled."
            )
        return val



"""Tool for generating an S matrix automatically from a Tidy3d simulation and modal port definitions."""

from __future__ import annotations

from typing import Optional

import autograd.numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.index import SimulationMap
from tidy3d.components.monitor import ModeMonitor
from tidy3d.components.source.field import ModeSource
from tidy3d.components.source.time import GaussianPulse
from tidy3d.components.types import Ax, Complex
from tidy3d.components.viz import add_ax_if_none, equal_aspect
from tidy3d.plugins.smatrix.ports.modal import Port
from tidy3d.plugins.smatrix.types import Element, MatrixIndex

from .base import FWIDTH_FRAC, AbstractComponentModeler


class ModalComponentModeler(AbstractComponentModeler):
    """A tool for modeling devices and computing scattering matrix elements.

    This class orchestrates the process of running multiple simulations to
    derive the scattering matrix (S-matrix) of a component. It uses modal
    sources and monitors defined by a set of ports.

    See Also
    --------
    **Notebooks**
        * `Computing the scattering matrix of a device <../../notebooks/SMatrix.html>`_
    """

    ports: tuple[Port, ...] = pd.Field(
        (),
        title="Ports",
        description="Collection of ports describing the scattering matrix elements. "
        "For each input mode, one simulation will be run with a modal source.",
    )

    run_only: Optional[tuple[MatrixIndex, ...]] = pd.Field(
        None,
        title="Run Only",
        description="Set of matrix indices that define the simulations to run. "
        "If ``None``, simulations will be run for all indices in the scattering matrix. "
        "If a tuple is given, simulations will be run only for the given matrix indices.",
    )

    element_mappings: tuple[tuple[Element, Element, Complex], ...] = pd.Field(
        (),
        title="Element Mappings",
        description="Tuple of S matrix element mappings, each described by a tuple of "
        "(input_element, output_element, coefficient), where the coefficient is the "
        "element_mapping coefficient describing the relationship between the input and output "
        "matrix element. If all elements of a given column of the scattering matrix are defined "
        "by ``element_mappings``, the simulation corresponding to this column is skipped automatically.",
    )

    @cached_property
    def sim_dict(self) -> SimulationMap:
        """Generates all :class:`.Simulation` objects for the S-matrix calculation.

        Returns
        -------
        Dict[str, Simulation]
            A dictionary where keys are task names and values are the
            corresponding :class:`.Simulation` objects. Each simulation is
            configured to excite a specific mode at a specific port and
            includes all necessary monitors.
        """

        sim_dict = {}
        mode_monitors = [self.to_monitor(port=port) for port in self.ports]

        for port_name, mode_index in self.matrix_indices_run_sim:
            port = self.get_port_by_name(port_name=port_name)

            port_source = self.shift_port(port=port)
            mode_source = self.to_source(port=port_source, mode_index=mode_index)

            new_mnts = list(self.simulation.monitors) + mode_monitors
            sim_copy = self.simulation.copy(update={"sources": [mode_source], "monitors": new_mnts})
            task_name = self.get_task_name(port=port, mode_index=mode_index)
            sim_dict[task_name] = sim_copy
        return SimulationMap(keys=tuple(sim_dict.keys()), values=tuple(sim_dict.values()))

    @cached_property
    def matrix_indices_monitor(self) -> tuple[MatrixIndex, ...]:
        """Returns a tuple of all possible matrix indices for monitoring.

        Each matrix index is a tuple of (port_name, mode_index).

        Returns
        -------
        Tuple[MatrixIndex, ...]
            A tuple of all possible matrix indices for the monitoring ports.
        """
        matrix_indices = []
        for port in self.ports:
            for mode_index in range(port.mode_spec.num_modes):
                matrix_indices.append((port.name, mode_index))
        return tuple(matrix_indices)

    @cached_property
    def matrix_indices_source(self) -> tuple[MatrixIndex, ...]:
        """Tuple of all the source matrix indices, which may be less than the total number of
        ports."""
        return super().matrix_indices_source

    @cached_property
    def matrix_indices_run_sim(self) -> tuple[MatrixIndex, ...]:
        """Tuple of all the matrix indices that will be used to run simulations."""
        return super().matrix_indices_run_sim

    @cached_property
    def port_names(self) -> tuple[list[str], list[str]]:
        """Returns lists of port names for inputs and outputs.

        Returns
        -------
        Tuple[List[str], List[str]]
            A tuple containing two lists: the first with the names of the
            output ports, and the second with the names of the input ports.
        """

        def get_port_names(matrix_elements: tuple[str, int]) -> list[str]:
            """Get the port names from a list of (port name, mode index)."""
            port_names = []
            for port_name, _ in matrix_elements:
                if port_name not in port_names:
                    port_names.append(port_name)
            return port_names

        port_names_in = get_port_names(self.matrix_indices_source)
        port_names_out = get_port_names(self.matrix_indices_monitor)

        return port_names_out, port_names_in

    def to_monitor(self, port: Port) -> ModeMonitor:
        """Creates a mode monitor from a given port.

        This monitor is used to measure the mode amplitudes at the port.

        Parameters
        ----------
        port : Port
            The port to convert into a monitor.

        Returns
        -------
        ModeMonitor
            A :class:`.ModeMonitor` configured to match the port's
            properties.
        """
        return ModeMonitor(
            center=port.center,
            size=port.size,
            freqs=self.freqs,
            mode_spec=port.mode_spec,
            name=port.name,
        )

    def to_source(
        self, port: Port, mode_index: int, num_freqs: int = 1, **kwargs
    ) -> list[ModeSource]:
        """Creates a mode source from a given port.

        This source is used to excite a specific mode at the port.

        Parameters
        ----------
        port : Port
            The port to convert into a source.
        mode_index : int
            The index of the mode to excite.
        num_freqs : int, optional
            The number of frequency points for the source, by default 1.

        Returns
        -------
        List[ModeSource]
            A list containing a single :class:`.ModeSource` configured to
            excite the specified mode at the port.
        """
        freq0 = np.mean(self.freqs)
        fdiff = max(self.freqs) - min(self.freqs)
        fwidth = max(fdiff, freq0 * FWIDTH_FRAC)
        return ModeSource(
            center=port.center,
            size=port.size,
            source_time=GaussianPulse(freq0=freq0, fwidth=fwidth),
            mode_spec=port.mode_spec,
            mode_index=mode_index,
            direction=port.direction,
            name=port.name,
            num_freqs=num_freqs,
            **kwargs,
        )

    def shift_port(self, port: Port) -> Port:
        """Generates a new port shifted slightly in the normal direction.

        This is to ensure that the source is placed just inside the
        simulation domain, away from the PML.

        Parameters
        ----------
        port : Port
            The port to shift.

        Returns
        -------
        Port
            A new :class:`.Port` object with its center shifted.
        """

        shift_value = self._shift_value_signed(port=port)
        center_shifted = list(port.center)
        center_shifted[port.size.index(0.0)] += shift_value
        port_shifted = port.copy(update={"center": center_shifted})
        return port_shifted

    @equal_aspect
    @add_ax_if_none
    def plot_sim(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
    ) -> Ax:
        """Plots the simulation with all sources added for troubleshooting.

        This method creates a temporary simulation with all mode sources
        activated to help visualize the setup.

        Parameters
        ----------
        x : float, optional
            The x-coordinate of the cross-section, by default None.
        y : float, optional
            The y-coordinate of the cross-section, by default None.
        z : float, optional
            The z-coordinate of the cross-section, by default None.
        ax : Ax, optional
            The matplotlib axes to plot on, by default None.

        Returns
        -------
        Ax
            The matplotlib axes with the plot.
        """

        plot_sources = []
        for port_source in self.ports:
            mode_source_0 = self.to_source(port=port_source, mode_index=0)
            plot_sources.append(mode_source_0)
        sim_plot = self.simulation.copy(update={"sources": plot_sources})
        return sim_plot.plot(x=x, y=y, z=z, ax=ax)

    @equal_aspect
    @add_ax_if_none
    def plot_sim_eps(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **kwargs,
    ) -> Ax:
        """Plots the permittivity of the simulation with all sources.

        This method is useful for visualizing the device geometry along
        with the placement of the sources.

        Parameters
        ----------
        x : float, optional
            The x-coordinate of the cross-section, by default None.
        y : float, optional
            The y-coordinate of the cross-section, by default None.
        z : float, optional
            The z-coordinate of the cross-section, by default None.
        ax : Ax, optional
            The matplotlib axes to plot on, by default None.
        **kwargs
            Additional keyword arguments passed to the plotter.

        Returns
        -------
        Ax
            The matplotlib axes with the plot.
        """

        plot_sources = []
        for port_source in self.ports:
            mode_source_0 = self.to_source(port=port_source, mode_index=0)
            plot_sources.append(mode_source_0)
        sim_plot = self.simulation.copy(update={"sources": plot_sources})
        return sim_plot.plot_eps(x=x, y=y, z=z, ax=ax, **kwargs)

    def _normalization_factor(self, port_source: Port, sim_data: SimulationData) -> complex:
        """Computes the normalization amplitude for the input mode.

        This is used to normalize the S-matrix elements.

        Parameters
        ----------
        port_source : Port
            The port that was excited.
        sim_data : SimulationData
            The data from the simulation run.

        Returns
        -------
        complex
            The complex amplitude of the input mode.
        """

        port_monitor_data = sim_data[port_source.name]
        mode_index = sim_data.simulation.sources[0].mode_index

        normalize_amps = port_monitor_data.amps.sel(
            f=np.array(self.freqs),
            direction=port_source.direction,
            mode_index=mode_index,
        )

        return normalize_amps.values

    @cached_property
    def max_mode_index(self) -> tuple[int, int]:
        """Returns the maximum mode indices for the in and out ports.

        Returns
        -------
        Tuple[int, int]
            A tuple containing the maximum mode index for the output ports
            and the maximum mode index for the input ports.
        """

        def get_max_mode_indices(matrix_elements: tuple[str, int]) -> int:
            """Get the maximum mode index for a list of (port name, mode index)."""
            return max(mode_index for _, mode_index in matrix_elements)

        max_mode_index_out = get_max_mode_indices(self.matrix_indices_monitor)
        max_mode_index_in = get_max_mode_indices(self.matrix_indices_source)

        return max_mode_index_out, max_mode_index_in
